"""Add job.type_filters

Revision ID: e81103413827
Revises: 47351f8c7ebc
Create Date: 2025-03-26 11:10:41.748248

"""
import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "e81103413827"
down_revision = "47351f8c7ebc"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("jobv2", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "type_filters", sa.JSON(), server_default="{}", nullable=False
            )
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("jobv2", schema=None) as batch_op:
        batch_op.drop_column("type_filters")

    # ### end Alembic commands ###
