"""Drop Task.default_args columns

Revision ID: f384e1c0cf5d
Revises: 4c308bcaea2b
Create Date: 2023-06-06 15:10:51.838607

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision = "f384e1c0cf5d"
down_revision = "4c308bcaea2b"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("task", "default_args")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "task", sa.Column("default_args", sqlite.JSON(), nullable=True)
    )
    # ### end Alembic commands ###
